package org.yasrun.game.kafunwars;

import java.util.ArrayList;
import java.util.List;

import org.yasrun.game.kafunwars.objects.Kafun;
import org.yasrun.game.kafunwars.objects.Laser;
import org.yasrun.game.kafunwars.objects.Nose;

import org.yasrun.game.kafunwars.R;
import android.content.Context;
import android.content.res.Resources;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.RectF;
import android.graphics.Paint.Style;
import android.view.MotionEvent;
import android.view.View;

/**
* 描画用のクラス
*/
public class KafunWarsView extends View implements Updatable {
    private final double DENSITY = getContext().getResources().getDisplayMetrics().density;

    private static final int KAFUN_MAX = 100;
    private static final int LASER_MAX = 100;
    private static final int POWER_MAX = 20;
    private static final int TIME_MAX = 100000;
    private static final float TIME_DELTA = 0.1F;
    private static final int NOSE_X = 75;
    private static final int NOSE_Y = 350;

    private final int GAUGE_X1 = adjustCoords(2);
    private final int GAUGE_Y1 = adjustCoords(40);
    private final int GAUGE_X2 = adjustCoords(7);
    private final int GAUGE_Y2 = adjustCoords(430);

    private Paint paintCyan = new Paint();
    private Paint paintGreen = new Paint();
    private Paint paintRed = new Paint();

    private Bitmap bmpHana = null;
    private Bitmap bmpKafun = null;

    private List<Kafun> kafuns = null;
    private List<Laser> lasers = null;

    private int touchStartX = 0;
    private int touchStartY = 0;
    private int power = 0;
    private int score = 0;
    private float time = 0;

    private Nose nose = null;

    private boolean gameOver = false;

    private UpdateHandler updateHandler = new UpdateHandler(this);

    /**
     * コンストラクタ
     *
     * @param c
     */
    public KafunWarsView(Context c) {
        super(c);
        setFocusable(true);

        final Resources res = this.getContext().getResources();
        bmpHana = BitmapFactory.decodeResource(res, R.drawable.hana);
        bmpKafun = BitmapFactory.decodeResource(res, R.drawable.kafun);

        paintCyan.setColor(Color.CYAN);
        paintCyan.setStyle(Style.FILL);
        paintGreen.setColor(Color.GREEN);
        paintGreen.setStyle(Style.STROKE);
        paintRed.setColor(Color.RED);
        paintRed.setStyle(Style.FILL);

        init();
    }

    private void init() {
        kafuns = new ArrayList<Kafun>();
        lasers = new ArrayList<Laser>();
        nose = new Nose(NOSE_X, NOSE_Y, 0, 0);
        nose.life = 30;
        gameOver = false;
    }

    /**
     * 処理開始
     */
    public void start() {
        update();
    }

    /**
     * タッチイベント
     */
    public boolean onTouchEvent(MotionEvent event) {

        if (gameOver || lasers.size() >= LASER_MAX) {
            return true;
        }

        switch (event.getAction()) {
        case MotionEvent.ACTION_DOWN:
            touchStartX = (int) (event.getX() / DENSITY);
            touchStartY = (int) (event.getY() / DENSITY);
            power = 1;
            break;
        case MotionEvent.ACTION_UP:
            int touchEndX = (int) (event.getX() / DENSITY);
            int touchEndY = (int) (event.getY() / DENSITY);
            lasers.add(new Laser(touchStartX - 4, touchStartY, touchEndX, touchEndY));
            lasers.add(new Laser(touchStartX, touchStartY, touchEndX, touchEndY));
            lasers.add(new Laser(touchStartX + 4, touchStartY, touchEndX, touchEndY));
            power = 0;
            break;
        case MotionEvent.ACTION_MOVE:
            if (power < POWER_MAX) {
                power++;
            }
            break;
        }

        return true;
    }

    /**
     * メイン
     */
    public void update() {
        if (!gameOver) {
            if (time < TIME_MAX) {
                time += TIME_DELTA;
            }

            if (Math.random() < 0.3 && kafuns.size() < KAFUN_MAX) {
                kafuns.add(new Kafun((int) (Math.random() * 320), 0, (int) (Math.random() * 8) - 4, 8));
            }

            isHit2();

            moveLasers();

            moveKafuns();
        }

        invalidate();

        updateHandler.sleep((int) (TIME_DELTA * 1000.0));
    }

    /**
     * レーザー移動
     */
    private void moveLasers() {
        final List<Laser> deadLasers = new ArrayList<Laser>();

        for (Laser laser : lasers) {
            isHit0(laser);
            laser.move();
            laser.judge();
            if (laser.isDead()) {
                deadLasers.add(laser);
            }
        }

        for (Laser laser : deadLasers) {
            lasers.remove(laser);
        }
    }

    /**
     * 判定1<br/>
     * 花粉とレーザーの当たり判定
     *
     * @param laser
     *            レーザー
     */
    private void isHit0(Laser laser) {
        for (Kafun kafun : kafuns) {
            if (!kafun.isDead() && laser.isInner(kafun.X + 4, kafun.Y + 4)) {
                kafun.setDead(true);
                score+=100;
            }
        }
    }

    /**
     * 花粉移動
     */
    private void moveKafuns() {
        final List<Kafun> deadKafuns = new ArrayList<Kafun>();

        for (Kafun kafun : kafuns) {
            if (isHit1(kafun)) {
                kafun.setDead(true);
                score+=100;
            }
            kafun.move();
            kafun.judge();
            if (kafun.life == -5) {
                deadKafuns.add(kafun);
            }
        }

        for (Kafun kafun : deadKafuns) {
            kafuns.remove(kafun);
        }
    }

    /**
     * 判定2<br/>
     * レーザーと花粉の当たり判定
     *
     * @param kafun
     *            判定対象花粉
     * @return レーザーに当たっているかどうか。trueならば当たっている。
     */
    private boolean isHit1(Kafun kafun) {
        boolean result = false;
        for (Laser laser : lasers) {
            if (!kafun.isDead() && laser.isInner(kafun.X + 4, kafun.Y + 4)) {
                result = true;
                break;
            }
        }
        return result;
    }

    /**
     * 判定3<br/>
     * 鼻と花粉の当たり判定
     */
    private void isHit2() {
        for (Kafun kafun : kafuns) {
            if (!kafun.isDead() && kafun.life >= 0
                    && kafun.X >= nose.X && kafun.X <= nose.X + bmpHana.getWidth() && kafun.Y >= nose.Y
                    && kafun.Y <= nose.Y + bmpHana.getHeight()) {
                kafun.life = -1;
                nose.life--;
                if (nose.life < 0) {
                    nose.life = 0;
                    gameOver = true;
                }
            }
        }
    }

    /**
     * 描画処理
     */
    protected void onDraw(Canvas canvas) {
        canvas.drawColor(Color.BLACK);

        // 枠
        canvas.drawRect(0, adjustCoords(4), adjustCoords(319), adjustCoords(479), paintGreen);

        // スコアと時間
        canvas.drawText("Score: " + Integer.toString(score), adjustCoords(20), adjustCoords(20), paintGreen);
        canvas.drawText("Time: " + String.format("%05d", (int) time), adjustCoords(220), adjustCoords(20), paintGreen);

        // ライフゲージ
        canvas.drawRect(GAUGE_X1,  GAUGE_Y1, GAUGE_X2, GAUGE_Y2, paintRed);
        canvas.drawRect(GAUGE_X1, adjustCoords(430 - nose.life * 13), GAUGE_X2, GAUGE_Y2, paintGreen);

        // 鼻
        canvas.drawBitmap(bmpHana, adjustCoords(nose.X), adjustCoords(nose.Y), paintCyan);

        // レーザーパワー
        if (power > 0) {
            int r = power / 2;
            canvas.drawOval(new RectF(adjustCoords(touchStartX - r), adjustCoords(touchStartY - r),
                    adjustCoords(touchStartX + r), adjustCoords(touchStartY + r)), paintCyan);
        }

        // 花粉
        for (Kafun kafun : kafuns) {
            if (!kafun.isDead()) {
                canvas.drawBitmap(bmpKafun, adjustCoords(kafun.X), adjustCoords(kafun.Y), paintCyan);
            } else {
                int x1 = adjustCoords(kafun.X);
                int y1 = adjustCoords(kafun.Y);
                int x2 = adjustCoords(kafun.X + 8);
                int y2 = adjustCoords(kafun.Y + 8);
                canvas.drawLine(x1, y1, x2, y2, paintCyan);
                canvas.drawLine(x2, y1, x1, y2, paintCyan);
            }
        }

        // レーザー
        for (Laser laser : lasers) {
            canvas.drawLine(adjustCoords(laser.X1), adjustCoords(laser.Y1), adjustCoords(laser.X2), adjustCoords(laser.Y2), paintCyan);
        }

        // ゲームオーバー時、「Game Over」を表示
        if (gameOver) {
            canvas.drawText("Game Over" ,adjustCoords(125), adjustCoords(200), paintGreen);
        }
    }

    /**
     * 座標を調節する
     * @param wh 座標値
     * @return 調節された座標
     */
    protected int adjustCoords(int wh) {
        return (int)(wh * DENSITY);
    }
}
